//============================================================================
// Name        : meepcsv_to_povray.cpp
// Author      : Emory R Stagmer (aka @VAXHeadroom)
// Version     :
// Copyright   : Copyright 2015 Emory R. Stagmer
// Description : Hello World in C++, Ansi-style
//============================================================================

#include <stdio.h>
#include <dirent.h>
#include <vector>
#include <string.h>
#include <errno.h>

using namespace std;
vector< dirent > files;
const char *arg_defs[]={"executable", "directory", "frame num", "x", "y", "z1", "z2", "z3", "z4"};

typedef struct{ char s[25]; } num_string;
typedef vector<num_string> number_line;
typedef vector<number_line> number_file;

int get_num( num_string &str, FILE *f, int &found_eoln )
    {
    int count = 0;
    int done = false;

    found_eoln = false;

    while(!done)
        {
        str.s[count] = fgetc(f);
        if(feof(f)) // end of file
            {
            done = true;
            continue;
            }
        if(str.s[count] == '\n' || str.s[count]=='\r' )
            {
            str.s[count] = 0x00;
            done = true;
            found_eoln = true;
            continue;
            }
        if( str.s[count]==',') // end of num
            {
            str.s[count] = 0x00;
            done = true;
            continue;
            }
        count++;
        }
//printf("<%s>\n", str.s);
    return(count);
    }

int process_files( FILE *x_file, FILE *y_file, FILE *z_file, FILE *output_file )
    {
    num_string sx, sy, sz;
    number_line lx, ly, lz;
    number_file fx, fy, fz;
    number_line::iterator lx_iter, ly_iter, lz_iter;
    number_file::iterator fx_iter, fy_iter, fz_iter;
    int done = false;
    int stat = true;
    int lenx, leny, lenz;
    int eolnx, eolny, eolnz;

    while(!done)
        {
        memset(&sx, 0x00, sizeof(sx)); // clear the string
        memset(&sy, 0x00, sizeof(sx));
        memset(&sz, 0x00, sizeof(sx));
        lenx = get_num(sx, x_file, eolnx); // get the next number from each
        leny = get_num(sy, y_file, eolny);
        lenz = get_num(sz, z_file, eolnz);
        done = feof(x_file) || feof(y_file) || feof(z_file);
        if(!done)
            {
            if(!eolnx && !eolny && !eolnz)
                {
                lx.push_back(sx); // push the string into the line
                ly.push_back(sy);
                lz.push_back(sz);
//                printf("c<%s,%s,%s> ", sx.s, sy.s, sz.s);
                }
            else
                {
                lx.push_back(sx); // push the string into the line
                ly.push_back(sy);
                lz.push_back(sz);
                fx.push_back(lx); // push the line into the file
                fy.push_back(ly);
                fz.push_back(lz);
                lx.clear(); // clear the line
                ly.clear();
                lz.clear();
//                printf("L\n");
                }
            }
        }

    // output sample
    //#declare max_rows = 250;
    //#declare max_cols = 250;
    //#declare x_vector = array[ 2 ][ 3 ]
    // {
    // {<1.1,1.2,1.3>,<2.1,2.2,2.3>,<3.1,3.2,3.3>},
    // {<4.1,4.2,4.3>,<5.1,5.2,5.3>,<6.1,6.2,6.3>}
    // }
    fprintf(output_file, "#declare max_cols = %d;\n", fx.front().size());
    fprintf(output_file, "#declare max_rows = %d;\n", fx.size());
    fprintf(output_file, "#declare csv_vector=array[max_rows][max_cols]\n");
    fprintf(output_file, "    {\n");

    for( fx_iter=fx.begin(), fy_iter=fy.begin(), fz_iter=fz.begin();
         fx_iter!=fx.end(), fy_iter!=fy.end(), fz_iter!=fz.end();
         fx_iter++, fy_iter++, fz_iter++ )
        {
        fprintf(output_file, "    { ");
        for( lx_iter=fx_iter->begin(), ly_iter=fy_iter->begin(), lz_iter=fz_iter->begin();
             lx_iter!=fx_iter->end(), ly_iter!=fy_iter->end(), lz_iter!=fz_iter->end();
             lx_iter++, ly_iter++, lz_iter++ )
            {
            fprintf(output_file, "<%s, %s, %s>, ", lx_iter->s, ly_iter->s, lz_iter->s);
            }
        fprintf(output_file, "}");
        if( &lx_iter->s != &fx_iter->back().s ) // if not the last element
            {
            fprintf(output_file, ","); // put a comma after the close brace
            }
        fprintf(output_file, "\n"); // end of line
        }

    fprintf(output_file, "    }\n");

    return(stat);
    }

int open_files( char *dir, char *frame, char e_or_h, char *base, char *which_slice )
    {
    int stat;
    char x_filename[1024];
    char y_filename[1024];
    char z_filename[1024];
    char output_filename[1024];

    FILE *x_file, *y_file, *z_file, *output_file;

    sprintf( x_filename, "%s%s-%cx%s.csv", dir, frame, e_or_h, base);
    sprintf( y_filename, "%s%s-%cy%s.csv", dir, frame, e_or_h, base);
    sprintf( z_filename, "%s%s-%cz%s.csv", dir, frame, e_or_h, base);
    sprintf( output_filename, "%s%s-%c_fields_%s.inc", dir, frame, e_or_h, which_slice);

    printf("%s\n", x_filename);
    printf("%s\n", y_filename);
    printf("%s\n", z_filename);
    printf("%s\n", output_filename);

    x_file = fopen(x_filename, "rt");
    if(!x_file)
        {
        printf("Can't open X file <%s>\n", x_filename);
        return(false);
        }
    y_file = fopen(y_filename, "rt");
    if(!y_file)
        {
        printf("Can't open Y file <%s>\n", y_filename);
        fclose(x_file);
        return(false);
        }
    z_file = fopen(z_filename, "rt");
    if(!z_file)
        {
        printf("Can't open Z file <%s>\n", z_filename);
        fclose(x_file);
        fclose(y_file);
        return(false);
        }
    output_file = fopen(output_filename, "wt");
    if(!output_file)
        {
        printf("Can't open output file <%s>\n", output_filename);
        fclose(x_file);
        fclose(y_file);
        fclose(z_file);
        return(false);
        }

    fprintf(output_file, "//Derived from the following files:\n");
    fprintf(output_file, "// X == <%s>\n", x_filename);
    fprintf(output_file, "// Y == <%s>\n", y_filename);
    fprintf(output_file, "// Z == <%s>\n", z_filename);

    stat = process_files(x_file, y_file, z_file, output_file);

    fclose(x_file);
    fclose(y_file);
    fclose(z_file);
    fclose(output_file);

    return(true);
    }

int main(int argc, char **argv)
    {
    DIR *dir_cntrl;
    dirent *dir_entry;
    int i;
    int stat;
    vector< dirent >::iterator dir_i;

    printf("MEEP CSV to POVRay converter.\n\n  Copyright 2015 Emory R. Stagmer\n  emory@untiedmusic.com\n  www.untiedmusic.com\n  @VAXHeadroom\n\n");
	printf("%d argument%c\n", argc, (argc==1?' ':'s'));
    for(i=0; i<argc && i<=8 ; i++)
        {
        printf("Arg[%d]:<%s> (%s)\n", i+1, argv[i], arg_defs[i]);
        }

    if( argc == 1 || argc > 9)
	    {
        if( argc > 9 )
            printf("-===- Invalid number of arguments, 9 max. -===-\n");
	    printf("\nUsage:\n");
	    printf(" note: Filenames are presumed to be ##-[e,h][x,y,z][x,y,z](optional-string).csv\n  i.e. 00-exx.csv, 00-hzy.csv, 00-eyz151.csv\n");
        printf("                                        ^^^ e(electric) or h(magnetic) is computed\n");
	    printf("                                              ^^^ this x,y,z is computed\n");
	    printf("                                                     ^^^ these characters are what you specify\n");
	    printf("                                                             ^^^ these characters are optional like the 151 above\n");
	    printf("Arguments:\n");
	    printf(" 1st: directory for CSV files - for current directory use \"./\"\n");
	    printf(" 2nd: n digit frame number - number of digits must match filename, i.e 00\n");
	    printf(" 3rd: base X filename - i.e. X for above example 00-ex\'X\'.csv\n");
        printf(" 4th: base Y filename - i.e. Y for above example 00-ex\'Y\'.csv\n");
        printf(" 5th: base Z filename - i.e. Z151 for above example 00-ex\'Z151\'.csv\n");
        printf(" 6th:(Optional) additional z filename\n");
        printf(" 7th:(Optional) additional z filename\n");
        printf(" 8th:(Optional) additional z filename\n");
        printf("Outputs:\n");
        printf(" E field filenames: e_fields_[x,y,z1,z2,z3,z4].inc\n");
        printf(" H field filenames: h_fields_[x,y,z1,z2,z3,z4].inc\n");
        return(1);
	    }

	dir_cntrl = opendir(argv[1]);
	if( dir_cntrl )
	    {
	    do
	        {
	        dir_entry = readdir(dir_cntrl);
	        if( dir_entry )
	            {
	            if( !strcmp("csv", &dir_entry->d_name[dir_entry->d_namlen-3]) ) // is it a CSV file?
	                {
	                if( !strncmp( argv[2], dir_entry->d_name, strlen(argv[2]) ) ) // does the frame number match?
	                    {
	                    files.push_back( *dir_entry ); // save it
	                    }
	                }
	            }
	        } while( dir_entry );
	    }
	else
	    {
	    i = errno;
	    printf("Directory could not be opened(errno=%d)\n", i);
	    return(2);
	    }

	printf("Found %d files\n", files.size());

	for( dir_i = files.begin(); dir_i != files.end(); dir_i++ )
	    {
	    printf("-- %s\n", dir_i->d_name);
	    }

	stat = open_files(argv[1], argv[2], 'e', argv[3], "X");
    stat = open_files(argv[1], argv[2], 'h', argv[3], "X");

    stat = open_files(argv[1], argv[2], 'e', argv[4], "Y");
    stat = open_files(argv[1], argv[2], 'h', argv[4], "Y");

    stat = open_files(argv[1], argv[2], 'e', argv[5], "Z1");
    stat = open_files(argv[1], argv[2], 'h', argv[5], "Z1");
    if( argc > 6 )
        {
        stat = open_files(argv[1], argv[2], 'e', argv[6], "Z2");
        stat = open_files(argv[1], argv[2], 'h', argv[6], "Z2");
        }
    if( argc > 7 )
        {
        stat = open_files(argv[1], argv[2], 'e', argv[7], "Z3");
        stat = open_files(argv[1], argv[2], 'h', argv[7], "Z3");
        }
    if( argc > 8 )
        {
        stat = open_files(argv[1], argv[2], 'e', argv[8], "Z4");
        stat = open_files(argv[1], argv[2], 'h', argv[8], "Z4");
        }

	return 0;
    }
